/*
 * sunlite.c
 *
 * Copyright (C) 2001-2004 Greg Kroah-Hartman (greg@kroah.com)
 * Copyright (C) Michael Stickel <michael@cubic.org>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c 
 * but has been rewritten to be easy to read and use, as no locks are now
 * needed anymore.
 *
 */

#include "dmxsunlite.h"

static int inputrate = 100;
module_param(inputrate, int, S_IRUGO);

DMXFamily *usbdmx2_family = NULL;

/* table of devices that work with this driver */
static struct usb_device_id usbdmx2_id_table [] =
  {
    { USB_DEVICE(USB_VENDOR_DIGITALARTSYSTEM, USB_PRODUCT_EZ_USB2) },  /* USBDMX2 */
    { USB_DEVICE(USB_VENDOR_DIGITALARTSYSTEM, USB_PRODUCT_EZ_USB2a) },  /* USBDMX2 firmware */
    { USB_DEVICE(USB_VENDOR_DIGITALARTSYSTEM, USB_PRODUCT_EZ_USB3) },  /* USBDMX-IN firmware */
    { USB_DEVICE(USB_VENDOR_DIGITALARTSYSTEM, USB_PRODUCT_EZ_USB3a) },  /* USBDMX-IN */
    { }                                         /* Terminating entry */
  };
MODULE_DEVICE_TABLE(usb, usbdmx2_id_table);


/* Get a minor range for your devices from the usb maintainer */
#define USB_DMX_MINOR_BASE	192


/* prevent races between write() and disconnect() */
static DECLARE_MUTEX (disconnect_sem);


static struct usb_driver dmx_driver;

#define USBDMX_IN_MAX_BULK_BUF_SIZE 0x400


/* Structure to hold all of our device specific stuff */
struct usb_dmx {
	struct usb_device *	udev;			/* the usb device for this device */
	struct usb_interface *	interface;		/* the interface for this device */
	__u8			bulk_in_endpointAddr;	/* the address of the bulk in endpoint */
	__u8			bulk_out_endpointAddr;	/* the address of the bulk out endpoint */
	struct kref		kref;

   struct urb* urb;

   DMXInterface             *dmx_interface;
   struct dmxusb_universe   *universe;


   atomic_t           urb_submit_pending;  /* value 1 during urb submit, reset in complete function*/

   unsigned char bulkin_buffer[USBDMX_IN_MAX_BULK_BUF_SIZE];
   unsigned char bulkout_buffer[1024];

   struct workqueue_struct *work_queue;
   struct work_struct work_st;

};
#define to_dmx_dev(d) container_of(d, struct usb_dmx, kref)

/*-----------------------------------------------------------
 *-- bulk_urb_complete
 *--
 *-- callback function for DMX read
 *---------------------------------------------------------*/

void bulk_urb_complete (struct urb *urb)
{
  unsigned char *ptmp_buffer;
  struct dmxusb_universe *u2d_u;
  struct usb_dmx *dev = (struct usb_dmx *)urb->context;
  int len;

  if (urb->status)
    info("urb->status: %d", urb->status);
  else if (urb->actual_length > USBDMX_IN_MAX_BULK_BUF_SIZE)
    info("urb->actual_length 0x%x > 0x%x", urb->actual_length, USBDMX_IN_MAX_BULK_BUF_SIZE);
  else if (!urb->context)
    info("no urb->context");
  else {
    ptmp_buffer = (unsigned char *)urb->transfer_buffer + 3; /* offset 3 */
    u2d_u = dev->universe;

    if (DMX_SLOTS<urb->actual_length)
      len=DMX_SLOTS;
    else
      len=urb->actual_length;

    if (u2d_u && u2d_u->universe) {
      if (memcmp(u2d_u->buffer, ptmp_buffer, len)) {
	//printk("received {$%02X,$%02X,$%02X,$%02X,$%02X}\n", ptmp_buffer[0],ptmp_buffer[1],ptmp_buffer[2],ptmp_buffer[3], ptmp_buffer[4]);

        memcpy(u2d_u->buffer, ptmp_buffer, len);
	
        u2d_u->data_avail=1;
        u2d_u->universe->signal_changed (u2d_u->universe, 0, len);

      };
    } else
      info("no universe");
  }

  atomic_set(&dev->urb_submit_pending, 0);

}



static void work_queue_callback(void *data) {
  int i, status;
  struct usb_dmx *dev = (struct usb_dmx *)data;

  if (!atomic_read(&dev->urb_submit_pending)) {    
    usb_fill_bulk_urb (dev->urb, dev->udev, usb_rcvbulkpipe(dev->udev, 2), &dev->bulkin_buffer, sizeof(dev->bulkin_buffer), bulk_urb_complete, dev);
    status=usb_submit_urb(dev->urb, GFP_ATOMIC);
    if (status)
      info("urb submit error: %d", status);
    else
      atomic_set(&dev->urb_submit_pending, 1);
  }


  i=queue_delayed_work(dev->work_queue, &dev->work_st, HZ/inputrate);

}


/*-----------------------------------------------------------
 *-- dmx_read_universe
 *--
 *-- Used to read a couple of slot-values to the universe.
 *---------------------------------------------------------*/
static int dmx_read_universe (DMXUniverse *u, off_t offs, DMXSlotType *buff, size_t size)
{

  if (u && size>0 && offs+size <= DMX_SLOTS) {
    struct dmxusb_universe *u2d_u = (struct dmxusb_universe *)u->user_data;

    if (u2d_u && u2d_u->dmx_if) {
      unsigned char *inbuffer = u2d_u->buffer;

      // printk(KERN_INFO "dmx_read_universe: data_avail=%d\n", u2d_u->data_avail);
      if (u2d_u->data_avail) {
        u2d_u->data_avail=0;

        //  printk(KERN_INFO "dmx_read_universe: data changed\n");

        if (offs+size>DMX_SLOTS)
          size = DMX_SLOTS-offs;

        memcpy (buff, inbuffer+offs, size);

        // printk(KERN_INFO "received {$%02X,$%02X,$%02X,$%02X}\n", inbuffer[0],inbuffer[1],inbuffer[2],inbuffer[3]);

        return size;
      }
      return 0;
    }
  }
  return -EINVAL;
}

/*-----------------------------------------------------------
 *-- dmx_write_universe
 *--
 *-- Used to write a couple of slot-values to the universe.
 *---------------------------------------------------------*/
static int dmx_write_universe (DMXUniverse *u, off_t offs, DMXSlotType *buff, size_t size)
{
  int i, retval=-EINVAL;

  /* prevent disconnects */
  down (&disconnect_sem);

  if (u && size>0 && offs+size <= 512) {
    struct dmxusb_universe *u2d_u = (struct dmxusb_universe *)u->user_data;
    if (u2d_u && u2d_u->dmx_if) {
	  
      for (i = 0; i < size; ++i)
        UDMX_SET_SLOT((char *) u2d_u->dmx_if->bulkout_buffer, offs+i, buff[i]);

      dmxusb_send_bulk (u2d_u->dmx_if, 0, u2d_u->dmx_if->bulkout_buffer, PACKET_SIZE);

      retval=0;
    }
  }

  up(&disconnect_sem);

  return retval;
}


/*-----------------------------------------------------------
 *-- usbdmx2_data_available
 *--
 *-- 
 *---------------------------------------------------------*/
int  usbdmx2_data_available (DMXUniverse *u, uint start, uint size)
{

  if (u && u->user_data)
    {
      struct dmxusb_universe *u2d_u = (struct dmxusb_universe *)u->user_data;
      if (u2d_u)
        return u2d_u->data_avail;
    }
  return 0;
}


/*-----------------------------------------------------------
 *-- usbdmx2_delete_universe
 *--
 *-- Delete the universe of a USBDMX2 interface.
 *---------------------------------------------------------*/
static int usbdmx2_delete_universe (DMXUniverse *u)
{
  printk ("usbdmx2_delete_universe called\n");
  if (u && u->user_data)
    {
      struct dmxusb_universe *u2d_u = (struct dmxusb_universe *)u->user_data;
      /* lock the universe */

      u2d_u->dmx_if->universe = NULL;
      u2d_u->dmx_if = NULL;
      u2d_u->universe = NULL;

      FREE(u->user_data);
      u->user_data = NULL;
      printk ("freeing universe->user_data done\n");
    }
  return 0;
}


/*-----------------------------------------------------------
 *-- usbdmx2_create_universe
 *--
 *-- Create the universe for the USBDMX2 interface.
 *---------------------------------------------------------*/
static int usbdmx2_create_universe (DMXUniverse *u, DMXPropList *pl)
{
  printk ("usbdmx2_create_universe called\n");
  if (u && u->interface) {
    struct dmxusb_universe *u2d_u = NULL;
      
    u2d_u = DMX_ALLOC(struct dmxusb_universe);
    if (u2d_u) {
      memset (u2d_u->buffer, 0, 512);

      u->user_data  = (void *)u2d_u;

      u->user_delete = usbdmx2_delete_universe;
      u->read_slots  = dmx_read_universe;
      u->data_available = usbdmx2_data_available;
      if (u->kind == 0) {
        /* output universe */
        u->write_slots = dmx_write_universe;
        strcpy (u->connector, "OUT");
        u->conn_id = 0;
        u2d_u->universe_id = 0;
      } else if (u->kind == 1) {
        /* input universe */
        strcpy (u->connector, "IN");
        u->conn_id = 1;
        u2d_u->universe_id = 0;
      } else
        return -1;

      u2d_u->dmx_if = (struct usb_dmx *)u->interface->user_data;
      u2d_u->universe    = u;
      u2d_u->dmx_if->universe = u2d_u;
    }
    return 0;
  }
  return -1;
}


/*-----------------------------------------------------------
 *-- usbdmx2_delete_interface
 *--
 *-- 
 *---------------------------------------------------------*/
static int usbdmx2_delete_interface (DMXInterface *dif)
{
  printk ("usbdmx2_delete_interface called\n");
  return 0;
}



/*-----------------------------------------------------------
 *-- usbdmx2_create_interface
 *--
 *-- Creates a dmx-interface for USBDMX2 hardware.
 *---------------------------------------------------------*/
static int usbdmx2_create_interface (DMXInterface *dif, DMXPropList *pl)
{
  printk ("usbdmx2_create_interface called\n");
  return 0;
}





static void dmx_delete(struct kref *kref)
{	
  struct usb_dmx *dev = to_dmx_dev(kref);

  if (dev->work_queue) {
    cancel_delayed_work(&dev->work_st);
    destroy_workqueue(dev->work_queue);
  }

  usb_put_dev(dev->udev);
  
  if (dev->urb) {
    usb_unlink_urb(dev->urb);
    usb_free_urb(dev->urb);
  }

  kfree (dev);
}

static int dmx_open(struct inode *inode, struct file *file)
{
  return -ENODEV;
}

static int dmx_release(struct inode *inode, struct file *file)
{
  return -ENODEV;
}

static ssize_t dmx_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
  return -EFAULT;
}


static ssize_t dmx_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
  return -ENOMEM;
}

static struct file_operations dmx_fops = {
  .owner =	THIS_MODULE,
  .read =		dmx_read,
  .write =	dmx_write,
  .open =		dmx_open,
  .release =	dmx_release,
};

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with devfs and the driver core
 */
static struct usb_class_driver dmx_class = {
  .name =		"usb/dmx%d",
  .fops =		&dmx_fops,
  .mode =		S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH,
  .minor_base =	USB_DMX_MINOR_BASE,
};

static int dmx_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
  struct usb_device *udev = interface_to_usbdev(interface);
  struct usb_dmx *dev = NULL;
  struct usb_host_interface *iface_desc;
  struct usb_endpoint_descriptor *endpoint;
  DMXDriver *drv = NULL;
  size_t buffer_size;
  int i;
  int retval = -ENOMEM;

  /*
   * Check if it is a known USBDMX2 release.
   */
  if (udev->descriptor.idVendor == USB_VENDOR_DIGITALARTSYSTEM) {
    if (udev->descriptor.idProduct == USB_PRODUCT_EZ_USB)
      info("Sunlite USBDMX1 found at address %d", udev->devnum);
    else if(udev->descriptor.idProduct == USB_PRODUCT_EZ_USB2 )
      info("Sunlite USBDMX2-firmware device found at address %d", udev->devnum);
    else if(udev->descriptor.idProduct == USB_PRODUCT_EZ_USB2a )
      info("Sunlite USBDMX2 device found at address %d", udev->devnum);
    else if(udev->descriptor.idProduct == USB_PRODUCT_EZ_USB3 )
      info("Sunlite USBDMX-IN-firmware device found at address %d", udev->devnum);
    else if(udev->descriptor.idProduct == USB_PRODUCT_EZ_USB3a )
      info("Sunlite USBDMX-IN device found at address %d", udev->devnum);
    else {
      warn(KERN_INFO "Sunlite USBDMX model not supported/tested. (id=%d)", udev->descriptor.idProduct);
      return retval;
    }
  } else
    return retval;


  /* allocate memory for our device state and initialize it */
  dev = kmalloc(sizeof(*dev), GFP_KERNEL);
  if (dev == NULL) {
    err("Out of memory");
    goto error;
  }
  memset(dev, 0x00, sizeof(*dev));
  kref_init(&dev->kref);

  dev->udev = usb_get_dev(interface_to_usbdev(interface));
  dev->interface = interface;

  /* save our data pointer in this interface device */
  usb_set_intfdata(interface, dev);

  /* we can register the device now, as it is ready */
  retval = usb_register_dev(interface, &dmx_class);
  if (retval) {
    /* something prevented us from registering this driver */
    err("Not able to get a minor for this device.");
    usb_set_intfdata(interface, NULL);
    goto error;
  }

  /* Upload the firmware if nessesary. */
  if ((udev->descriptor.idVendor == USB_VENDOR_DIGITALARTSYSTEM) &&
      (udev->descriptor.idProduct == USB_PRODUCT_EZ_USB2 || udev->descriptor.idProduct == USB_PRODUCT_EZ_USB3 ) ) {
    int err;
    info (DRIVERNAME " loading firmware...");
    err = USBDMX2_startup (udev, udev->descriptor.idProduct);
    if (err)
      info (DRIVERNAME "...failed");
    else
      info (DRIVERNAME "...succeded");

    // we want to manage the interface
    return 0;
  }


  /* set up the endpoint information */
  /* use only the first bulk-in and bulk-out endpoints */
  iface_desc = interface->cur_altsetting;
  for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
    endpoint = &iface_desc->endpoint[i].desc;

    if (!dev->bulk_in_endpointAddr &&
        (endpoint->bEndpointAddress & USB_DIR_IN) &&
        ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
         == USB_ENDPOINT_XFER_BULK)) {
      /* we found a bulk in endpoint */
      buffer_size = endpoint->wMaxPacketSize;
      dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
    }

    if (!dev->bulk_out_endpointAddr &&
        !(endpoint->bEndpointAddress & USB_DIR_IN) &&
        ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
         == USB_ENDPOINT_XFER_BULK)) {
      /* we found a bulk out endpoint */
      dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
    }
  }

  if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
    err("Could not find both bulk-in and bulk-out endpoints");
    goto error;
  }

  if(udev->descriptor.idProduct == USB_PRODUCT_EZ_USB2a) {
    printk ("initize bulkout-buffer.\n");
    udmx_init_packet (dev->bulkout_buffer);
  }

  if(udev->descriptor.idProduct == USB_PRODUCT_EZ_USB3a) {
    printk ("initize USBDMX-IN hardware\n");
    udmx_in_init (dev);
  }

  /* USBDMX-IN */
  if (udev->descriptor.idProduct == USB_PRODUCT_EZ_USB3a) { 

    dev->urb = usb_alloc_urb(0, GFP_KERNEL);

    dev->work_queue = create_singlethread_workqueue("dmxwq");

    INIT_WORK(&dev->work_st, work_queue_callback, dev);

    queue_delayed_work(dev->work_queue, &dev->work_st, HZ/inputrate);
    info("starting work queue, inputrate: %d\n", inputrate);
  }


  /* let the user know what node this device is now attached to */
  info("USB DMX device now attached to USBDMX-%d (minor)", interface->minor);


  /*
   * Create a USBDMX2 specific record and a DMX universe.
   */
  drv = dmx_find_driver (usbdmx2_family, "usbdmx2");
  if (drv) {

    DMXInterface *dmx_if = drv->create_interface (drv, dmxproplist_vacreate ("usbdev=%l", (long)udev));

    dmx_if->user_data = (void *)dev;
    dmx_if->user_delete = usbdmx2_delete_interface;

    if (dmx_if && dmx_if->user_data) {

      DMXUniverse  *uout=NULL;
     
      dev->universe = NULL;
      dev->dmx_interface = dmx_if;
      dev->udev = udev; /* usb-device handle */

      if (udev->descriptor.idProduct == USB_PRODUCT_EZ_USB3a)
        uout = dmx_if->create_universe (dmx_if, 1, NULL); /* input */
      if (udev->descriptor.idProduct == USB_PRODUCT_EZ_USB2a || udev->descriptor.idProduct == USB_PRODUCT_EZ_USB)
        uout = dmx_if->create_universe (dmx_if, 0, NULL); /* output */

      if (uout)	{
        printk("input/output-universes created\n");

        return 0;
      }

      dmx_if->delete(dmx_if);
    }
    
    return 0;

  } else
    printk (KERN_INFO "unable to find driver for dmx-family usbdmx2.usbdmx2\n");


 error:
  if (dev)
    kref_put(&dev->kref, dmx_delete);
  return retval;

}

static void dmx_disconnect(struct usb_interface *interface)
{
  struct usb_dmx *dev;
  int minor = interface->minor;

  /* prevent write from racing dmx_disconnect() */
  down (&disconnect_sem);

  dev = usb_get_intfdata(interface);

  if (dev->dmx_interface) {
    printk (KERN_INFO "delete usbdmx2 interface\n");
    if (dev->dmx_interface->delete)
      dev->dmx_interface->delete (dev->dmx_interface);
    /* does a cascaded delete on the universes of that interface */
  }

  usb_set_intfdata(interface, NULL);

  /* give back our minor */
  usb_deregister_dev(interface, &dmx_class);

  /* decrement our usage count */
  kref_put(&dev->kref, dmx_delete);

  up(&disconnect_sem);

  info("USB DMX #%d now disconnected", minor);
}

static struct usb_driver dmx_driver = {
  .owner =	THIS_MODULE,
  .name =	"USBDMX",
  .probe =	dmx_probe,
  .disconnect =	dmx_disconnect,
  .id_table =	usbdmx2_id_table,
};

static int __init usb_dmx_init(void)
{
  int result;
  DMXDriver *drv = NULL;

  usbdmx2_family = dmx_create_family("usbdmx2");
  if (!usbdmx2_family) {
    printk (KERN_INFO "unable to register dmx-family USBDMX2\n");
    return -1;
  }

  drv = usbdmx2_family->create_driver (usbdmx2_family, "usbdmx2", usbdmx2_create_universe, NULL);
  if (!drv) {
    usbdmx2_family->delete(usbdmx2_family, 0);
    info("failed to create usbdmx2 driver\n");
    return -1;
  }

  drv->num_out_universes = 1;
  drv->num_in_universes = 1;
  drv->user_create_interface = usbdmx2_create_interface;

  /* register this driver with the USB subsystem */
  result = usb_register(&dmx_driver);
  if (result) {
    usbdmx2_family->delete(usbdmx2_family, 0);
    err("usb_register failed. Error number %d", result);
  } else
    info("USBDMX2 registered.");

  return result;
}

static void __exit usb_dmx_exit(void)
{
  /* deregister this driver with the USB subsystem */
  usb_deregister(&dmx_driver);

  mdelay(5);
  if (usbdmx2_family)
    usbdmx2_family->delete(usbdmx2_family, 0);

}

module_init (usb_dmx_init);
module_exit (usb_dmx_exit);

MODULE_AUTHOR("(c) 2001 Michael Stickel <michael@cubic.org> http://llg.cubic.org");
MODULE_DESCRIPTION("Driver for the Sunlite-DMX512 interface");
MODULE_LICENSE("GPL");

